import pickle
import string

from abc import abstractmethod
from enum import Enum
from nltk.tokenize import TweetTokenizer
from transformers import BertTokenizer


class TokenType(Enum):
    WORD = 0
    CHARACTER = 1
    CHAR = 1
    WORDPIECE = 2
    WP = 2

    @classmethod
    def from_str(cls, s):
        s = s.lower()
        try:
            return cls(int(s))
        except ValueError:
            pass
        if s in ['word']:
            return cls.WORD
        elif s in ['char', 'character']:
            return cls.CHAR
        elif s in ['wordpiece', 'wp']:
            return cls.WP
        else:
            raise ValueError("{} not recognized as a token type."
                             " Available token types are 'word',"
                             " 'char' and 'wordpiece'.".format(s))


class BaseTokenizer:
    """ This is not a class that should be instantiated. Its encode() method
        calls the tokenize() method which is not declared, but should be
        implemented in subclasses.
    """

    def __init__(self, vocab_fname=None):
        """ Args:
                vocab_fname (str): optional. If given, the vocabulary will be
                loaded from the file, which should be formatted with one token
                per line.
        """
        self.pad_token = '[PAD]'
        self.unk_token = '[UNK]'
        self.pad_token_id = 0
        self.unk_token_id = 1
        if vocab_fname is not None:
            with open(vocab_fname, 'rb') as f:
                vocab = pickle.load(f)
            self.init_vocab(vocab)

    def init_vocab(self, vocab):
        self.vocab = {word: (number + 2)
                      for number, word in enumerate(vocab)}
        self.vocab[self.unk_token] = self.unk_token_id
        self.vocab[self.pad_token] = self.pad_token_id
        self.vocab_size = len(self.vocab)
        self.ids_to_tokens = {v: k for k, v in self.vocab.items()}

    @abstractmethod
    def tokenize(self):
        raise NotImplementedError

    @abstractmethod
    def detokenize(self):
        raise NotImplementedError

    def encode(self, text):
        if type(text) == str:
            text = self.tokenize(text)
        try:
            encoded = self._encode_decode(text,
                                          converter=self.vocab,
                                          default_value=self.unk_token_id)
        except AttributeError as e1:
            e2 = AttributeError("It seems vocabulary was not initialized and"
                                " therefore encoding could not be performed.")
            e2.__traceback__ = e1.__traceback__
            raise e2
        return encoded

    def decode(self, encoded_text):
        # Unlike the use of the _encode_decode() in encode(), this does not
        # need to be in a try / except block because if this function is
        # called, encode() must have been called beforehand.
        decoded = self._encode_decode(encoded_text,
                                      converter=self.ids_to_tokens,
                                      default_value=self.unk_token)
        decoded = self.detokenize(decoded)
        return decoded

    def _encode_decode(self, input_txt, converter, default_value):
        output_text = []
        for token in input_txt:
            try:
                output_text.append(converter[token])
            except KeyError:
                output_text.append(default_value)
        return output_text


class CharTokenizer(BaseTokenizer):
    """ Character-level tokenizer.
    """

    def __init__(self, vocab_fname):
        super().__init__(vocab_fname)
        self.tokenization = TokenType.CHAR

    def tokenize(self, text):
        return list(text)

    def detokenize(self, tokenized_txt):
        return ''.join(tokenized_txt)


class WordTokenizer(BaseTokenizer):
    """ Word-level tokenizer, uses NLTK's tweet tokenizer to separate words.
    """

    def __init__(self, vocab_fname):
        super().__init__(vocab_fname)
        self.tokenization = TokenType.WORD
        self.tknzr = TweetTokenizer()

    def tokenize(self, text):
        return self.tknzr.tokenize(text)

    def detokenize(self, tokenized_txt):
        detok_txt = ''
        for tok in tokenized_txt:
            detok_txt += tok
            if tok not in string.punctuation:
                detok_txt += ' '
        return detok_txt


def get_tokenizer(tokenization, vocab):
    if type(tokenization) == str:
        tokenization = TokenType.from_str(tokenization)
    if tokenization == TokenType.CHAR:
        return CharTokenizer(vocab)
    elif tokenization == TokenType.WP:
        tok = BertTokenizer.from_pretrained(vocab)
        tok.tokenization = TokenType.WP
        return tok
    else:
        return WordTokenizer(vocab)



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pubtatortool",
    version="0.6.4",
    author="Guilhem Piat",
    author_email="guilhem.piat@protonmail.com",
    description="A package for loading and manipulating PubTator files "
                "as Python objects.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gpiat/PubTator-Parsing-Tool",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: Public Domain",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'nltk',
        'diff_match_patch',
        'transformers',
    ],
)



import itertools
import nltk
from diff_match_patch import diff_match_patch


def text_preprocess(text):
    """ Takes the raw title and abstract as written in the file
        and keeps only the relevant text and PMID.
        Args:
            text (str): raw text of the form "PMID|?|Text\n"
                with '?' being 't' or 'a' depending on whether
                the text is a title or abstract.
        return:
            pmid (str): PMID of the article, of 1 to 8 digits
            text (str): cleaned text
    """
    # separating PMID and letter from the text.
    # *text captures the rest of the list in case
    # there happens to be a | in the text.
    pmid, _, *text = text.split("|")
    # joining the various fields of the title in case there
    # happens to be a | in the title text; removing \n
    text = "|".join(text)[:-1]

    return pmid, text


class UMLS_Entity:
    """ This data structure allows easy access to the various fields
        which describe a mention of a UMLS entity in a text.
    """

    def __init__(self, text):
        self._orig_text = text
        # The first field that is ignored is the PMID of the document.
        (_, self.start_idx, self.stop_idx, self.mention_text,
         self.semantic_type_ID, self.cui) = text.split('\t')

        # The CUI may be formatted as "C0123456" or "UMLS:C0123456".
        # We want to chop off the "UMLS:" if applicable.
        self.cui = self.cui.split(':')[-1]

        self.start_idx = int(self.start_idx)
        self.stop_idx = int(self.stop_idx)

    def __str__(self):
        return self._orig_text


class PubTatorDocument:
    """ This class instantiates a document from the PubTator corpus
        using the information provided in the PubTator format.
        Attr:
            abstract (str): Abstract of the article
            char_level_targets
            pmid (str): PMID of the document
            raw_text (str): simple concatenation of title and abstract. The
                indexing of characters in raw_text matches the one used in
                PubTator entity mention annotations.
            sentences (list<list<str>>): List of sentences, each being a list
                of tokens.
            sent_start_end_indices (list<(int, int)>): List of
                (start_index, end_index) pairs for each sentence (the index
                refers to the character indices in the original text string)
            start_end_indices
            targets
            text (list<str>): title fused with abstract but as a list of tokens
            title (str): Title of the article
            token_to_char_lookup
            tokenization
            tokenizer
            umls_entities (list<UMLS_Entity>): list of UMLS entity
                mentions in the text
            vocab
    """

    def __init__(self, title, abstract, mentions, tokenizer):
        """ Args:
                - (str) title: raw title line of text
                - (str) abstract: raw abstract line of text
                - (list<str>) mentions: list of raw lines
                    of text containing the UMLS entity mentions.
                - tokenizer: TODO.
        """
        self.pmid, self.title = text_preprocess(title)
        _, self.abstract = text_preprocess(abstract)
        # no space is insterted between title and abstract to match up
        # with PubTator PubTator format.
        self.raw_text = self.title + '\n' + self.abstract
        self.umls_entities = [UMLS_Entity(entity) for entity in mentions]
        self.tokenizer = tokenizer

        try:
            sentence_tok = nltk.data.load('tokenizers/punkt/english.pickle')
        except LookupError:
            nltk.download('punkt')
            sentence_tok = nltk.data.load('tokenizers/punkt/english.pickle')

        # We get a list of (start_idx, end_idx) pairs for each sentence
        self.sent_start_end_indices = list(
            sentence_tok.span_tokenize(self.raw_text))

        # the \n at the end of the title is a sentence delimeter which
        # the sentence tokenizer does not recognize. Therefore we manually
        # split the appropriate span if need be.
        newline = self.raw_text.index('\n')
        for i, (start, end) in enumerate(self.sent_start_end_indices):
            if start < newline + 1 < end:
                self.sent_start_end_indices = (
                    self.sent_start_end_indices[:i] +
                    [(start, newline),
                    (newline + 1, end)] +
                    self.sent_start_end_indices[i + 1:])
                break

        self.sentences = [self.title]
        self.sentences.extend(sentence_tok.tokenize(self.abstract))
        self.sentences = [self.tokenizer.tokenize(sentence)
                          for sentence in self.sentences]
        # for example, if tokenization is wordpiece, self.sentences is now
        # of the form [['Title'], ['Abs', '##tract', '.'], ['Stuff', '.']]

        self.text = self.tokenizer.tokenize(self.raw_text)
        self.tokenization = tokenizer.tokenization
        self._initialize_targets()

        # list of all start and end indices of all entities
        # originally the stop index is exclusive, but we need it
        # to be inclusive and vice-versa for the start index.
        self.start_end_indices = list(itertools.chain(
            [(e.start_idx - 1, e.stop_idx - 1) for e in self.umls_entities]))

    def _initialize_targets(self):
        char_level_targets = [None] * len(self.raw_text)
        for i in range(len(char_level_targets)):
            for e in self.umls_entities:
                if i >= e.start_idx and i < e.stop_idx:
                    char_level_targets[i] = e.cui, e.semantic_type_ID
                    continue
                elif i > e.stop_idx:
                    continue

        token = iter(self.text)
        concat_tokens = ''.join(self.text)
        dmp = diff_match_patch()
        diff = dmp.diff_main(self.raw_text, concat_tokens)
        # this diff library creates diffs of the form:
        #   [(flag, substring), (flag, substring), ...]
        # where "flag" can be 1, -1 or 0 depending on whether
        # the substring is in concat_tokens but not raw_text,
        # vice-versa, or in both repectively.
        # example:
        #   raw_text = 'Nonylphenol'
        #   concat_tokens = 'Non##yl##phe##no##l'
        #   diff = [(0, 'Non'), (1, '##'), (0, 'yl'), (1, '##'),
        #           (0, 'phe'), (1, '##'), (0, 'no'), (1, '##'), (0, 'l')]
        # However, it is much easier to handle a character-level diff, e.g. :
        #   [(0, 'N'), (0, 'o'), (0, 'n'), (1, '#'), (1, '#'), (0, 'y'), ...]
        # This is what we set out to do with the following list
        # comprehension, which may seem obscure, but it is 30%
        # faster than the equivalent loop, which can be written as:
        # new_diff = []
        # for flag, sub_str in diff:
        #     new_diff += list(zip([flag] * len(sub_str), sub_str))
        # diff = new_diff
        #
        # diff = list(itertools.chain(*[zip([a] * len(b), b)
        #                               for a, b in diff]))
        #
        # In actuality, once we have a sequence of numbers that tell us
        # whether a character comes from one string, the other or both, the
        # character itself is redundant information. Removing it from
        # consideration further divides execution time by 2.
        diff = list(itertools.chain(*[[flag] * len(sub_str)
                                      for flag, sub_str in diff]))

        # initializing the targets for each token as None
        token_targets = [None] * len(self.text)
        # This helps us keep track of where we are within the current token
        # so that we know when to move on to the next token
        chars_left_in_current_token = len(next(token))
        # Index of the currently tracked character in the raw_text, which
        # always refers (if possible) to the same character as the character
        # defined by chars_left_in_current_token. I'm not completely certain
        # why it needs to start at -1 rather than 0, it just works.
        current_char_index = -1
        current_token_index = 0
        # we maintain a lookup table of tokens to characters for the purpose
        # of versatility.
        self.token_to_char_lookup = {current_token_index: []}
        for flag in diff:
            # hypothetically, we can choose any time to assign the
            # token's label but we choose to do so when we're on the
            # last character of the current token.
            if chars_left_in_current_token == 0:
                token_targets[current_token_index] =\
                    char_level_targets[current_char_index]
                current_token_index += 1
                # We also add the current token to the lookup table
                self.token_to_char_lookup[current_token_index] = []
                chars_left_in_current_token = len(next(token))
            # we're pointing to the same character in both the
            # raw_text and the tokenized text, so update both
            # pointers for the next iteration
            if flag == 0:
                current_char_index += 1
                # we add the current character index to the lookup table
                self.token_to_char_lookup[current_token_index].append(
                    current_char_index)
                chars_left_in_current_token -= 1
            # we're on a character missing in the tokenized text,
            # so we skip it
            elif flag == -1:
                current_char_index += 1
                # we do not add the current character index to the lookup
                # table, since it is not in the tokens
            # we're on a character missing in the raw_text, so we skip it
            else:
                chars_left_in_current_token -= 1
        self.char_level_targets = char_level_targets
        self.targets = token_targets

    def get_vocab(self):
        try:
            return self.vocab
        except AttributeError:
            self.vocab = list(set(self.text))
        return self.vocab

    def write_to(self, filename):
        with open(filename, 'a') as f:
            f.write('|'.join([self.pmid, 't', self.title]) + '\n')
            f.write('|'.join([self.pmid, 'a', self.abstract]) + '\n')
            for entity in self.umls_entities:
                f.write(str(entity) + '\n')
            f.write('\n')

    def sentence_targets(self):
        """ The targets are a single list with one element per (sub)token.
            In case the targets are needed and must be formatted the same
            as sentences, this generator will extract the relevant sublists
            on-the-fly.
            This assumes that the sentence parser doesn't split two sentences
            in the middle of a (sub)token.
        """
        start = 0
        for sentence in self.sentences:
            end = start + len(sentence)
            yield self.targets[start:end]
            start = end



from .pubtator_document import PubTatorDocument
from .tokenization import WordTokenizer


def peek(file, n=1):
    """ Reads n lines ahead in a file without changing the file
        object's current position in the file
        Args:
            - file: a text file
            - n (int): how many lines ahead should be read, defaults to 1
        Return:
            - line: last line read
    """
    pos = file.tell()
    lines = [file.readline() for i in range(n)]
    file.seek(pos)
    return lines[-1]


class PubTatorCorpus:
    """ This class instantiates a PubTator corpus using one or more
        PubTator-formatted files. Its main purpose is to iterate over
        documents with the documents() generator function.
        Attributes:
            tokenizer: A tokenizer that implements a `tokenize()`
                method and has a `tokenization` attribute of type
                `tokenization_tools.TokenType`.
            tokenization: a TokenType (the same as its tokenizer)
            n_documents: The number of documents in the corpus
            document_list: a list containing all of the documents as
                PubTatorDocument objects.
            cuids: a dict with all of the CUIs that appear in the corpus
                as keys, and their number of occurences as values
            stids: a dict with all of the STIs that appear in the corpus
                as keys, and their number of occurences as values
            vocab: a set of all tokens used in the corpus
            nconcepts: total number of CUIs in the corpus.
    """

    def __init__(self, fnames, tokenizer=None):
        """ Args:
                - fnames (list<str>): list of filenames in the corpus
                - auto_looping (bool): whether retrieving documents should
                    automatically loop or not
                - tokenizer: Optional (defaults to
                    `tokenization_tools.WordTokenizer`). A tokenizer that
                    implements a `tokenize()` method and has a `tokenization`
                    attribute of type `tokenization_tools.TokenType`.
        """

        self._filenames = fnames
        self._currentfile = 0

        self.tokenizer = tokenizer or WordTokenizer(None)
        self.tokenization = self.tokenizer.tokenization
        self._init_documents()
        self.n_documents = len(self.document_list)

        self.cuids, self.stids, self.vocab = self._get_cuids_and_vocab()
        self.nconcepts = len(self.cuids)

    def _get_cuids_and_vocab(self):
        """ Collects the CUIDs, STIDs and vocabulary of the corpus.
            Should not be used outside of the constructor, because
            it relies on the document counter being at the start.
            If you need CUIDs or vocab, use the appropriate attributes.
        """
        cuids = {}
        stids = {}
        vocab = set()
        for document in self.document_list:
            for entity in document.umls_entities:
                if entity.cui in cuids:
                    cuids[entity.cui] += 1
                else:
                    cuids[entity.cui] = 1
                if entity.semantic_type_ID in stids:
                    stids[entity.semantic_type_ID] += 1
                else:
                    stids[entity.semantic_type_ID] = 1
            vocab = vocab.union(document.get_vocab())
        return cuids, stids, vocab

    def _init_documents(self):
        self.document_list = []
        while self._currentfile < len(self._filenames):
            # opening the file -- not using `with` because
            # it causes excessive indentation
            f = open(self._filenames[self._currentfile], 'r')

            next_line = None
            while next_line != '' and peek(f) != '':
                title = f.readline()
                abstract = f.readline()
                next_line = f.readline()

                # after the abstract, each entity mention is written on a
                # separate line. The next document comes after a newline.
                umls_entities = []
                while next_line != '\n' and next_line != '':
                    # [:-1] deletes the trailing newline character
                    umls_entities.append(next_line[:-1])
                    next_line = f.readline()

                self.document_list.append(PubTatorDocument(title, abstract,
                                                           umls_entities,
                                                           self.tokenizer))
            f.close()

            self._currentfile += 1

    def documents(self):
        """ Yields a document containing:
                - pmid (str): the next document's PMID
                - title (str): the next document's title
                - abstract (str): the next document's abstract
                - umls_entities (list<str>): list of UMLS entities
                    for the next document
        """
        for document in self.document_list:
            yield document
